Object subclass: #DbmConnection
	instanceVariableNames: 'cookie '
	classVariableNames: 'OpenConnections '
	poolDictionaries: ''
	category: 'ESUG-DBM'!


!DbmConnection methodsFor: 'closing'!

close
	OpenConnections removeKey: self cookie.
	^self basicClose! !

!DbmConnection methodsFor: 'accessing'!

fetch: key
	<primitive: 10003>
	Object notFoundSignal raise!

store: key value: value
	<primitive: 10004>
	self error: 'store:value failed with code ' , self failureCode printString!

store: key value: value flags: flags
	<primitive: 10004>
	self error: 'store:value:flags: failed with code ' , self failureCode printString! !

!DbmConnection methodsFor: 'removing'!

delete: key
	<primitive: 10005>
	Object notFoundSignal raise! !

!DbmConnection methodsFor: 'enumerating'!

do: aBlock
	| key |
	key := self firstKey.
	key notNil ifTrue: [aBlock value: key].
	[(key := self nextKey) isNil]
		whileFalse: [aBlock value: key].
	^nil!

firstKey
	<primitive: 10006>
	self error: 'firstKey failed with code ' , self failureCode printString!

nextKey
	<primitive: 10007>
	self error: 'nextKey failed with code ' , self failureCode printString! !

!DbmConnection methodsFor: 'converting'!

asDictionary
	| dict |
	dict := Dictionary new.
	self do: [ :key | dict at: key put: (self fetch: key)].
	^dict! !

!DbmConnection methodsFor: 'finalization'!

finalize
	"my clone has expired."
	Transcript cr; show: 'Finalizing DBM connection #' , cookie printString.
	self basicClose! !

!DbmConnection methodsFor: 'private'!

basicClose
	<primitive: 10002>!

cookie
	^cookie!

failureCode
	<primitive: 9999>! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

DbmConnection class
	instanceVariableNames: ''!


!DbmConnection class methodsFor: 'class initialization'!

initialize
	(ObjectMemory dependents includes: self)
		ifFalse: [ObjectMemory addDependent: self].
	OpenConnections isNil
		ifTrue: 	[OpenConnections := WeakDictionary new]! !

!DbmConnection class methodsFor: 'instance creation'!

open: aDatabase
	| connection |
	connection := self primOpen: aDatabase.
	OpenConnections at: connection cookie put: connection.
	^connection!

primOpen: aDatabase
	<primitive: 10001 errorCode: errorValue>
	errorValue halt.
	self error: 'open: failed with code ' self failureCode printString!

qprimOpen: aDatabase
	<primitive: 10001>
	self error: 'open: failed with code ' self failureCode printString! !

!DbmConnection class methodsFor: 'snapshot'!

returnFromSnapshot
	OpenConnections isNil ifFalse:
		[OpenConnections keys do:
			[ :key | OpenConnections removeKey: key]]!

update: aspect with: parameter from: sender 
	(sender == ObjectMemory and: [aspect == #returnFromSnapshot])
		ifTrue: [self returnFromSnapshot]! !

!DbmConnection class methodsFor: 'private'!

failureCode
	<primitive: 9999>! !

!DbmConnection class methodsFor: 'examples'!

aliases
	| dict dbm key |
	dict := Dictionary new.
	dbm := DbmConnection open: '/etc/aliases'.
	key := dbm firstKey.
	[key isNil] whileFalse:
		[dict at: key put: (dbm fetch: key).
		key := dbm nextKey].
	^dict!

db
	| dict dbm key |
	dict := Dictionary new.
	dbm := DbmConnection open: 'db'.
	key := dbm firstKey.
	[key isNil] whileFalse:
		[dict at: key put: (dbm fetch: key).
		key := dbm nextKey].
	^dict! !
DbmConnection initialize!

