#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <ndbm.h>

#include "userprim.h"

#define MIN(X,Y)	((X) < (Y) ? (X) : (Y))
#define MAXLEN		256


#define UPEdbmCouldNotOpen	(-32)
#define UPEdbmKeyNotFound	(-33)
#define UPEdbmCouldNotStore	(-33)


char *versionString= "Objectworks\\Smalltalk(R), Release 4.1f (DBM)";
char *resourceClassName = "St80";


DBM *getDbm(dbm_conn)
upHandle dbm_conn;
{
  return (DBM *)UPSTtoCint(UPinstVarAt(dbm_conn, 1));
}


static datum stringToDatum(string)
upHandle string;
{
  upInt size;
  static upChar cString[MAXLEN];
  datum dtm;

  UPmustBeString(string);
  size= UPcopySTtoCstring(string, cString, MAXLEN, 1);
  cString[size]= '\0';
  fprintf(stderr, "CONVERTED STRING <%s>, strlen= %d, size= %d\n",
	  cString, strlen(cString), size);
  dtm.dptr= cString;
  dtm.dsize= size;
  return dtm;
}


static upHandle datumToString(dtm)
datum dtm;
{
  upHandle string= UPallocString(' ', dtm.dsize);
  fprintf(stderr, "CONVERTING TO ST STRING SIZE %d\n", dtm.dsize);
  UPcopyCtoSTstring(string, dtm.dptr, dtm.dsize, 1);
  return string;
}


void dbmOpen(recv, name)
upHandle recv, name;
{
  DBM *db;
  upChar cname[MAXLEN];
  upInt size;
  upHandle instance;

  UPmustBeString(name);
  size= UPcopySTtoCstring(name, cname, MAXLEN, 1);
  cname[size]= '\0';
  if (!(db= dbm_open(cname, O_RDWR | O_CREAT, 0644)))
    UPfail(UPEdbmCouldNotOpen);
  instance= UPallocFsObject(recv);
  UPinstVarAtPut(instance, 1, UPCtoSTint((int)db));
  fprintf(stderr, "ALLOCATING DBM HANDLE %d\n", (int)db);
  UPreturnHandle(instance);
}


void dbmClose(recv)
upHandle recv;
{
  DBM *db= getDbm(recv);
  fprintf(stderr, "CLOSING DBM HANDLE %d\n", (int)db);
  dbm_close(db);
  UPreturnNil();
}


void dbmFetch(recv, key)
upHandle recv, key;
{
  DBM *db;
  datum dbmkey, dbmvalue;
  upInt size;
  
  db= getDbm(recv);
  fprintf(stderr, "FETCHING DBM HANDLE %d\n", (int)db);
  dbmkey= stringToDatum(key);
  if ((dbmvalue= dbm_fetch(db, dbmkey)).dptr) {
    fprintf(stderr, "GOT DATUM WITH SIZE %d\n", dbmvalue.dsize);
    fprintf(stderr, "               1st CHAR <%c>\n", dbmvalue.dptr[0]);
    UPreturnHandle(datumToString(dbmvalue));
  }
  UPfail(UPEdbmKeyNotFound);
}


void dbmStore(recv, key, value, flags)
upHandle recv, key, value, flags;
{
  DBM *db;
  datum ckey, cvalue;
  int status;

  db= getDbm(recv);
  ckey= stringToDatum(key);
  ckey.dptr= strdup(ckey.dptr);
  cvalue= stringToDatum(value);
  status= dbm_store(db, ckey, cvalue, DBM_REPLACE);
  free(ckey.dptr);
  if (status < 0)
    UPfail(UPEdbmCouldNotStore);
  UPreturnHandle(value);
}


void dbmDelete(recv, key)
upHandle recv, key;
{
  DBM *db;
  datum ckey, cvalue;

  db= getDbm(recv);
  ckey= stringToDatum(key);
  if (dbm_delete(db, ckey) < 0)
    UPfail(UPEdbmKeyNotFound);
  UPreturnHandle(recv);
}


void dbmFirstKey(recv)
upHandle recv;
{
  DBM *db;
  datum ckey;
  upHandle string;

  db= getDbm(recv);
  if ((ckey= dbm_firstkey(db)).dptr) {
    string= datumToString(ckey);
    UPreturnHandle(string);
  }
  UPreturnNil();
}


void dbmNextKey(recv)
upHandle recv;
{
  DBM *db;
  datum ckey;
  upHandle string;

  db= getDbm(recv);
  if ((ckey= dbm_nextkey(db)).dptr) {
    string= datumToString(ckey);
    UPreturnHandle(string);
  }
  UPreturnNil();
}


char *UPinstall()
{
    UPaddPrimitive(10001, dbmOpen, 1);
    UPaddPrimitive(10002, dbmClose, 0);
    UPaddPrimitive(10003, dbmFetch, 1);
    UPaddPrimitive(10004, dbmStore, 2);
    UPaddPrimitive(10005, dbmDelete, 1);
    UPaddPrimitive(10006, dbmFirstKey, 0);
    UPaddPrimitive(10007, dbmNextKey, 0);
    return "dbm primitives";
}


UPInitLinkRegistry() {}


main(argc, argv)
char **argv;
{
  return st80Main(argc, argv, (char **)0);
}
